#!/usr/bin/env python
"""
GOLUM-CNP v1.0

Copyright Laboratoire des Sciences du Climat et de l'Environnement (LSCE)
          Unite mixte CEA-CNRS-UVSQ 

Contact person:
 yilong.wang@lsce.ipsl.fr

Reference:
 Y. Wang, P. Ciais, D. Goll, Y. Huang, et al.,
    GOLUM-CNP v1.0: a data-driven modeling of carbon, 
    nitrogen and phosphorus cycles in major terrestrial biomes.
 Geosci. Model Dev. Discuss., https://doi.org/10.5194/gmd-2018-23,
 in review, 2018 

History:
Version   Date          Comment
1.0       Jan 2018      initial version
""" 
# 0. load modules and functions
import numpy as np
from netCDF4 import Dataset
# home-made modules and functions (see corresponding .py files)
from ncload import ncload
import ss_Anal

path=''  # work directory

# 1. read variables
# All the data have been regridded at the resolution of 0.25 degree from the original products
# Please refer to the manuscript for the detailed information of references
landnc=ncload(path+'Land')      # load netcdf file of land/sea mask
land=landnc.get('land')         # read land/sea mask
# read biome map
# biome 0: TRF; biome 1: TEDF; biome 2: TECF; biome 3: BOCF; 
# biome 4: TUN; biome 5: C3 grass; biome 6: C4 grass
PFTnc=ncload(path+'PFTmap/MainPFT.nc')
PFT=PFTnc.get('PFTindex')
carbnc=ncload(path+'CARDAMOM.nc')       # load netcdf file of CARDAMOM
# read CARDAMOM C variables (mean value), Bloom et al., 2016
NPP,LEAF,ROOT,WOOD,LITTER,SOM,RT_LEAF,RT_ROOT,RT_WOOD,RT_LITTER,RT_SOM,GPP,\
 AF_FOL,AF_ROOT,AF_WOOD,ETAC,FRAC_Fire_LEAF,FRAC_Fire_ROOT,FRAC_Fire_WOOD,FRAC_Fire_LITTER,FRAC_Fire_SOM\
  =carbnc.get('Mean_FL_NPP','Mean_CP_FOL','Mean_CP_ROO','Mean_CP_WOO','Mean_CP_LIT','Mean_CP_SOM',\
              'Mean_RT_FOL','Mean_RT_ROO','Mean_RT_WOO','Mean_RT_LIT','Mean_RT_SOM','Mean_FL_GPP',\
              'Mean_AF_FOL','Mean_AF_ROO','Mean_AF_WOO','Mean_ETAC',\
              'Mean_FRAC_Fire_FOL','Mean_FRAC_Fire_ROO','Mean_FRAC_Fire_WOO','Mean_FRAC_Fire_LIT','Mean_FRAC_Fire_SOM')
# read standard deviation of CARDAMOM C variables, Bloom et al., 2016
std_NPP,std_LEAF,std_ROOT,std_WOOD,std_LITTER,std_SOM,\
 std_RT_LEAF,std_RT_ROOT,std_RT_WOOD,std_RT_LITTER,std_RT_SOM,std_GPP,\
 std_AF_FOL,std_AF_ROOT,std_AF_WOOD,std_ETAC,\
 std_FRAC_Fire_LEAF,std_FRAC_Fire_ROOT,std_FRAC_Fire_WOOD,std_FRAC_Fire_LITTER,std_FRAC_Fire_SOM\
  =carbnc.get('STD_FL_NPP','STD_CP_FOL','STD_CP_ROO','STD_CP_WOO','STD_CP_LIT','STD_CP_SOM',\
              'STD_RT_FOL','STD_RT_ROO','STD_RT_WOO','STD_RT_LIT','STD_RT_SOM','STD_FL_GPP',\
              'STD_AF_FOL','STD_AF_ROO','STD_AF_WOO','STD_ETAC',\
              'STD_FRAC_Fire_FOL','STD_FRAC_Fire_ROO','STD_FRAC_Fire_WOO','STD_FRAC_Fire_LIT','STD_FRAC_Fire_SOM')
deponc=ncload(path+'NP_deposition.nc')   # load netcdf file of N/P deposition
Nd,Pd=deponc.get('N_dep','P_dep')        # read gridded map of N/P deposition, Wang et al., 2017
Nfixnc=ncload(path+'Nfix.nc')            # load netcdf file of BNF
Nfix=Nfixnc.get('Nfix')                  # read gridded map of BNF, Peng et al., submitted to GBC
# read fraction of N gas loss (denitrification) to total N loss (denitrification + leaching), which is derived from a global dataset of 15N measurements in soils, Goll et al., 2017
gfracnc=ncload(path+'fgas.nc')
gfrac=gfracnc.get('fgas_N')
Pwnc=ncload(path+'Pw.nc')                # load netcdf file of P release from rock weathering
Pw=Pwnc.get('Pw')                        # read fluxes of P release from rock weathering, Hartmann et al., 2014
Leachingnc=ncload(path+'LeachingRatio_GRDCrunoff_ERASW.nc')   # load netcdf file of leaching ratio
# read leaching ratio, derived from GRDC runoff data and ERA-Interim/Land soil water content
LeachingRatio=Leachingnc.get('LeachingRatio')
# read fraction of labile soil P to total soil P, Yang and Post, 2011
DissfracPnc=ncload(path+'DissFracP.nc')
Dissfrac,Dissfrac_unc=DissfracPnc.get('Frac','unc')
# N stoichiometries, Zechmeister-Boltenstern et al., 2015
# s1: foliage; s2: root; s3: wood; s5: SOM
# eps1: resorption coefficient in foliage; 
# eps2: resorption coefficient in root; 
# eps3: resorption coefficient in wood
# biome 0: TRF; biome 1: TEDF; biome 2: TECF; biome 3: BOCF; 
# biome 4: TUN; biome 5: C3 grass; biome 6: C4 grass
s1={0:1./25. *14./12.,1:1./25. *14./12.,2:1./59. *14./12.,3:1./49. *14./12.,\
    4:1./39. *14./12.,5:1./25. *14./12.,6:1./49. *14./12.}
s2={0:1./47. *14./12.,1:1./59. *14./12.,2:1./67. *14./12.,3:1./57. *14./12.,\
    4:1./39. *14./12.,5:1./88. *14./12.,6:1./54. *14./12.}
s3={0:1./148.*14./12.,1:1./471.*14./12.,2:1./844.*14./12.,3:1./525.*14./12.,\
    4:0.     *14./12.,5:0.     *14./12.,6:0      *14./12.}
s5={0:1./16. *14./12.,1:1./19. *14./12.,2:1./20. *14./12.,3:1./32. *14./12.,\
    4:1./25. *14./12.,5:1./10. *14./12.,6:1./13. *14./12.}
eps1={0:1.-25./49,1:1.-25./65,2:1.-59./87.5,3:1.-49./56,4:1.-39./85,5:1.-25./63,6:1.-49./71}
eps2={0:0.,1:0.,2:0.,3:0.,4:0.,5:0.,6:0.}
eps3={0:0.,1:0.,2:0.,3:0.,4:0.,5:0.,6:0.}
# P stoichiometries
# l1: foliage; l2: root; l3: wood; l5: SOM
# tht1: resorption coefficient in foliage; 
# tht2: resorption coefficient in root; 
# tht3: resorption coefficient in wood
# biome 0: TRF; biome 1: TEDF; biome 2: TECF; biome 3: BOCF; 
# biome 4: TUN; biome 5: C3 grass; biome 6: C4 grass
l1={0:1./1027. *31./12.,1:1./867.  *31./12.,2:1./1232. *31./12.,3:1./1049. *31./12.,\
    4:1./753.  *31./12.,5:1./1278. *31./12.,6:1./2167. *31./12.}
l2={0:1./3125. *31./12.,1:1./1962. *31./12.,2:1./1186. *31./12.,3:1./1574. *31./12.,\
    4:1./1300. *31./12.,5:1./2829. *31./12.,6:1./1300. *31./12.}
l3={0:1./13574.*31./12.,1:1./11179.*31./12.,2:1./24297.*31./12.,3:1./19734.*31./12.,\
    4:0.       *31./12.,5:0.       *31./12.,6:0        *31./12.}
l5={0:1./159.  *31./12.,1:1./366.  *31./12.,2:1./302.  *31./12.,3:1./960.  *31./12.,\
    4:1./509.  *31./12.,5:1./130.  *31./12.,6:1./138.  *31./12.}
tht1={0:1.-1027./3924,1:1.-867./2715,2:1.-1232./2350,3:1.-1049./2692,4:1.-753./2441,5:1.-1278./5065,6:1.-2167./3262}
tht2={0:0.,1:0.,2:0.,3:0.,4:0.,5:0.,6:0.}
tht3={0:0.,1:0.,2:0.,3:0.,4:0.,5:0.,6:0.}

# 2. prepare 
# identity matrix to compute tangent linear of the model
X_tl=np.diag(np.ones(39))
# create netcdf file to store the results
resultnc=Dataset(path+'results_ss.nc','w')
# create dimensions
resultnc.createDimension('latitude',580)
resultnc.createDimension('longitude',1440)
resultnc.createDimension('ninput',39)
resultnc.createDimension('noutput',45)
# create variables
# latitudinal range of the model is from 90N to 55S; 
# this variable store the latitude of the center of grid cells
varlat=resultnc.createVariable('latitude','f',('latitude',))
varlat[:]=np.arange(89.875,-55.,-0.25)
varlat.units='degrees_north'
# longitudinal range of the model is from 180W to 180E; 
# this variable store the latitude of the center of grid cells
varlon=resultnc.createVariable('longitude','f',('longitude',))
varlon[:]=np.arange(-179.875,180.,0.25)
varlon.units='degrees_east'
# variable for all input data
varin=resultnc.createVariable('inputs','f',('ninput','latitude','longitude'))
# variable for all output data
varout=resultnc.createVariable('outputs','f',('noutput','latitude','longitude'))
# variable for uncertainty in input data
varin_unc=resultnc.createVariable('inputs_unc','f',('ninput','latitude','longitude'))
# variable for uncertainty in output data
varout_unc=resultnc.createVariable('outputs_unc','f',('noutput','latitude','longitude'))
# variable for sensitivity matrix
varsen=resultnc.createVariable('sensitivity','f',('noutput','ninput','latitude','longitude'))
# default value are set to -999
varin[:]=-999.;varin.missing_value=-999.
varout[:]=-999.;varout.missing_value=-999.
varin_unc[:]=-999.;varin_unc.missing_value=-999.
varout_unc[:]=-999.;varout_unc.missing_value=-999.
varsen[:]=-999.;varsen.missing_value=-999.

# 3. run the model; the model is run for each land grid cell
for ii in range(49,580):  # loop through latitudes for land 
  for jj in range(1440):  # loog through latitudes
    if Pw[ii,jj]>-1:Pw_pix=Pw[ii,jj]/1000.   # kg km-2 a-1 -> g m-2 a-1
    else:Pw_pix=0.        # if no value are provided for P release from rock weathering, then 0 is assumed
    if NPP[ii,jj]>0:      # where NPP>0, i.e. exclude desert and ice 
      if PFT[ii,jj]<4:    # tree biomes
        # put all input data in one array
        X=np.array([GPP[ii,jj],AF_FOL[ii,jj],AF_ROOT[ii,jj],AF_WOOD[ii,jj],ETAC[ii,jj],\
                    RT_LEAF[ii,jj],RT_ROOT[ii,jj],RT_WOOD[ii,jj],\
                    RT_LITTER[ii,jj],RT_SOM[ii,jj],\
                    FRAC_Fire_LEAF[ii,jj],FRAC_Fire_ROOT[ii,jj],FRAC_Fire_WOOD[ii,jj],\
                    FRAC_Fire_LITTER[ii,jj],FRAC_Fire_SOM[ii,jj],\
                    s1[PFT[ii,jj]],s2[PFT[ii,jj]],s3[PFT[ii,jj]],\
                    s5[PFT[ii,jj]],\
                    eps1[PFT[ii,jj]],eps2[PFT[ii,jj]],eps3[PFT[ii,jj]],\
                    Nd[ii,jj],Nfix[ii,jj],\
                    LeachingRatio[ii,jj],gfrac[ii,jj]/100.,\
                    l1[PFT[ii,jj]],l2[PFT[ii,jj]],l3[PFT[ii,jj]],\
                    l5[PFT[ii,jj]],\
                    tht1[PFT[ii,jj]],tht2[PFT[ii,jj]],tht3[PFT[ii,jj]],\
                    Pw_pix,Pd[ii,jj],Dissfrac[ii,jj],1./25.,NPP[ii,jj],0.75])
        # uncertainty (10% for variables whose uncertainties were not reported in original literature)
        B=np.diag([std_GPP[ii,jj],std_AF_FOL[ii,jj],std_AF_ROOT[ii,jj],std_AF_WOOD[ii,jj],std_ETAC[ii,jj],\
                   std_RT_LEAF[ii,jj],std_RT_ROOT[ii,jj],std_RT_WOOD[ii,jj],\
                   std_RT_LITTER[ii,jj],std_RT_SOM[ii,jj],\
                   std_FRAC_Fire_LEAF[ii,jj],std_FRAC_Fire_ROOT[ii,jj],std_FRAC_Fire_WOOD[ii,jj],\
                   std_FRAC_Fire_LITTER[ii,jj],std_FRAC_Fire_SOM[ii,jj]]\
                  +[0.1*X[i] for i in range(15,22)]\
                  +[Nd[ii,jj]*0.6]\
                  +[0.1*X[i] for i in range(23,34)]\
                  +[Pd[ii,jj]]+[Dissfrac_unc[ii,jj]]+[X[-3]*0.1]\
                  +[std_NPP[ii,jj],0.075])
        # run the model for tree biomes
        y,H,sigma=ss_Anal.tree(X,X_tl,B**2) # B**2: variances
        # write input data to netcdf file
        varin[:,ii,jj]=X
        # write uncertainty in input data to netcdf file
        varin_unc[:,ii,jj]=np.diag(B)
        # write output data to netcdf file
        varout[:,ii,jj]=y
        # write uncertainty in output data to netcdf file
        varout_unc[:,ii,jj]=np.sqrt(np.diag(sigma))
        # write sensitivity matrix to netcdf file
        varsen[:,:,ii,jj]=H
      else:               # grass biomes   
        # put all input data in one array
        X=np.array([GPP[ii,jj],AF_FOL[ii,jj]+AF_WOOD[ii,jj]/2.,AF_ROOT[ii,jj]+AF_WOOD[ii,jj]/2.,0.,\
                    ETAC[ii,jj],RT_LEAF[ii,jj],RT_ROOT[ii,jj],RT_WOOD[ii,jj],\
                    RT_LITTER[ii,jj],RT_SOM[ii,jj],\
                    FRAC_Fire_LEAF[ii,jj],FRAC_Fire_ROOT[ii,jj],FRAC_Fire_WOOD[ii,jj],\
                    FRAC_Fire_LITTER[ii,jj],FRAC_Fire_SOM[ii,jj],\
                    s1[PFT[ii,jj]],s2[PFT[ii,jj]],s3[PFT[ii,jj]],\
                    s5[PFT[ii,jj]],\
                    eps1[PFT[ii,jj]],eps2[PFT[ii,jj]],eps3[PFT[ii,jj]],\
                    Nd[ii,jj],Nfix[ii,jj],\
                    LeachingRatio[ii,jj],gfrac[ii,jj]/100.,\
                    l1[PFT[ii,jj]],l2[PFT[ii,jj]],l3[PFT[ii,jj]],\
                    l5[PFT[ii,jj]],\
                    tht1[PFT[ii,jj]],tht2[PFT[ii,jj]],tht3[PFT[ii,jj]],\
                    Pw_pix,Pd[ii,jj],Dissfrac[ii,jj],1./25.,NPP[ii,jj],0.75])
        # uncertainty (10% for variables whose uncertainties were not reported in original literature)
        B=np.diag([std_GPP[ii,jj],np.sqrt(std_AF_FOL[ii,jj]**2+std_AF_WOOD[ii,jj]**2/4.),\
                   np.sqrt(std_AF_ROOT[ii,jj]**2+std_AF_WOOD[ii,jj]**2/4.),0.,std_ETAC[ii,jj],\
                   std_RT_LEAF[ii,jj],std_RT_ROOT[ii,jj],std_RT_WOOD[ii,jj],\
                   std_RT_LITTER[ii,jj],std_RT_SOM[ii,jj],\
                   std_FRAC_Fire_LEAF[ii,jj],std_FRAC_Fire_ROOT[ii,jj],std_FRAC_Fire_WOOD[ii,jj],\
                   std_FRAC_Fire_LITTER[ii,jj],std_FRAC_Fire_SOM[ii,jj]]\
                  +[0.1*X[i] for i in range(15,22)]\
                  +[Nd[ii,jj]*0.6]\
                  +[0.1*X[i] for i in range(23,34)]\
                  +[Pd[ii,jj]]+[Dissfrac_unc[ii,jj]]+[X[-3]*0.1]\
                  +[std_NPP[ii,jj],0.075])
        # run the model for grass biomes
        y,H,sigma=ss_Anal.grass(X,X_tl,B**2) # B**2: variances
        # write input data to netcdf file
        varin[:,ii,jj]=X
        # write uncertainty in input data to netcdf file
        varin_unc[:,ii,jj]=np.diag(B)
        # write output data to netcdf file
        varout[:,ii,jj]=y
        # write uncertainty in output data to netcdf file
        varout_unc[:,ii,jj]=np.sqrt(np.diag(sigma))
        # write sensitivity matrix to netcdf file
        varsen[:,:,ii,jj]=H

# 4. finalize: close netcdf file
resultnc.close()
