#!/usr/bin/env python
"""
GOLUM-CNP v1.0

Copyright Laboratoire des Sciences du Climat et de l'Environnement (LSCE)
          Unite mixte CEA-CNRS-UVSQ 

Contact person:
 yilong.wang@lsce.ipsl.fr

Reference:
 Y. Wang, P. Ciais, D. Goll, Y. Huang, et al.,
    GOLUM-CNP v1.0: a data-driven modeling of carbon, 
    nitrogen and phosphorus cycles in major terrestrial biomes.
 Geosci. Model Dev. Discuss., https://doi.org/10.5194/gmd-2018-23,
 in review, 2018 

History:
Version   Date          Comment
1.0       Jan 2018      initial version
"""
import numpy as np

"""
In this code, we use GPP * GPP allocation fractions intead of NPP * NPP allocation fractions, 
since GPP and GPP allocation fractions are the direct variables that CARDAMOM provides.
In CARDAMOM, the autotrophic respiration is substracted from GPP first, then the remaining GPP (NPP)
are allocated to foliage, root and wood. With this simplification of using GPP and GPP allocation
fractions, we can use direct variables from CARDAMOM, ratehr than to compute NPP allocation 
fractions from CARDAMOM before running GOLUM-CNP.
"""

# Function to compute tree grid cells
def tree(X,X_tl,B): 
"""
 X   : input data (mean value)
 X_tl: ancillary variable to compute tagent linear of the code
 B   : variances of input data
"""
  # read mean values from input
  # gpp    : GPP; 
  # gam1-3 : GPP allocation fraction to foliage, root, wood (= NPP allocation fraction*NPP/GPP)
  # etac   : fraction of litter outflux that enters SOM
  # tau1-5 : residence times of foliage, root, wood, litter, SOM
  # f1-5   : fraction of fire to total outgoing flux from foliage, root, wood, litter, SOM
  # s1-5   : N:C ratio of foliage, root, wood, SOM
  # eps1-3 : N resportion coefficients of foliage, root (=0), wood (=0)
  # Nd     : N deposition
  # Nfix   : BNF
  # leachingratio :  leaching ratio (runoff/soil-water-content)
  # gfrac  : fraction of denitrification to the total loss of N (denitrification + leaching)
  # l1-5   : P:C ratio of foliage, root, wood, SOM
  # tht1-3 : P resportion coefficients of foliage, root (=0), wood (=0)
  # Pw     : release of P from rock weathering
  # Pd     : P deposition
  # dissfrac : fraction of labile soil P that is dissolved in the soil water
  # fsorb  : transform rate from labile soil P to occluded P
  # npp    : NPP
  # fashP  : fraction of P retained in the residual ash during a fire (=0.75) 
  gpp,gam1,gam2,gam3,etac=X[:5]
  tau1,tau2,tau3,tau4,tau5=X[5:10]
  f1,f2,f3,f4,f5=X[10:15]
  s1,s2,s3,s5=X[15:19]
  eps1,eps2,eps3=X[19:22]
  Nd,Nfix=X[22:24]
  leachingratio=X[24]
  gfrac=X[25]
  l1,l2,l3,l5=X[26:30]
  tht1,tht2,tht3=X[30:33]
  Pw,Pd=X[33:35]
  dissfrac=X[35]
  fsorb=X[36]
  npp=X[37]
  fashP=X[38]
  # ancillary variable for tangent-linear computation of the code
  gpp_tl,gam1_tl,gam2_tl,gam3_tl,etac_tl=X_tl[:5]
  tau1_tl,tau2_tl,tau3_tl,tau4_tl,tau5_tl=X_tl[5:10]
  f1_tl,f2_tl,f3_tl,f4_tl,f5_tl=X_tl[10:15]
  s1_tl,s2_tl,s3_tl,s5_tl=X_tl[15:19]
  eps1_tl,eps2_tl,eps3_tl=X_tl[19:22]
  Nd_tl,Nfix_tl=X_tl[22:24]
  leachingratio_tl=X_tl[24]
  gfrac_tl=X_tl[25]
  l1_tl,l2_tl,l3_tl,l5_tl=X_tl[26:30]
  tht1_tl,tht2_tl,tht3_tl=X_tl[30:33]
  Pw_tl,Pd_tl=X_tl[33:35]
  dissfrac_tl=X_tl[35]
  fsorb_tl=X_tl[36]
  npp_tl=X_tl[37] 
  fashP_tl=X_tl[38]

  # 1. C cycle
  # Steady-state (SS) foliage C pool (Eq. E1)
  C1=gpp*gam1*tau1
  C1_tl=gpp_tl*gam1*tau1+gpp*gam1_tl*tau1+gpp*gam1*tau1_tl
  # SS root C pool (Eq. E2)
  C2=gpp*gam2*tau2
  C2_tl=gpp_tl*gam2*tau2+gpp*gam2_tl*tau2+gpp*gam2*tau2_tl
  # SS wood C pool (Eq. E3)
  C3=gpp*gam3*tau3
  C3_tl=gpp_tl*gam3*tau1+gpp*gam3_tl*tau3+gpp*gam3*tau3_tl
  # SS litter C pool (Eq. E4)
  C4=gpp*(gam1*(1.-f1)+gam2*(1.-f2))*tau4
  C4_tl=gpp_tl*(gam1*(1.-f1)+gam2*(1.-f2))*tau4+gpp*(gam1_tl*(1.-f1)-gam1*f1_tl+gam2_tl*(1.-f2)-gam2*f2_tl)*tau4+gpp*(gam1+gam2)*tau4_tl
  # SS SOM C pool (Eq. E5) (gpp*gam3=C3/tau3)
  C5=(gpp*gam3*(1.-f3)+etac*C4/tau4)*tau5
  C5_tl=(gpp_tl*gam3*(1.-f3)+gpp*gam3_tl*(1.-f3)-gpp*gam3*f3_tl+(etac_tl*C4+etac*C4_tl)/tau4-etac*C4/tau4**2*tau4_tl)*tau5+(gpp*gam3+etac*C4/tau4)*tau5_tl
  # NPP
  Fc=npp
  Fc_tl=npp_tl
  # mortality of foliage -> litter
  mort1=gpp*gam1*(1.-f1)
  mort1_tl=(gpp_tl*gam1+gpp*gam1_tl)*(1.-f1)-gpp*gam1*f1_tl
  # mortality of root -> litter
  mort2=gpp*gam2*(1.-f2)
  mort2_tl=(gpp_tl*gam2+gpp*gam2_tl)*(1.-f2)-gpp*gam2*f2_tl
  # mortality of wood -> SOM
  mort3=gpp*gam3*(1.-f3)
  mort3_tl=(gpp_tl*gam3+gpp*gam3_tl)*(1.-f3)-gpp*gam3*f3_tl
  # fire flux from foliage -> atmosphere
  fire1=gpp*gam1*f1
  fire1_tl=gpp_tl*gam1*f1+gpp*gam1_tl*f1+gpp*gam1*f1_tl
  # fire flux from root -> atmosphere
  fire2=gpp*gam2*f2
  fire2_tl=gpp_tl*gam2*f2+gpp*gam2_tl*f2+gpp*gam2*f2_tl
  # fire flux from wood -> atmosphere
  fire3=gpp*gam3*f3
  fire3_tl=gpp_tl*gam3*f3+gpp*gam3_tl*f3+gpp*gam3*f3_tl
  # fire flux from litter -> atmosphere
  fire4=C4/tau4*f4
  fire4_tl=C4_tl/tau4*f4-fire4/tau4*tau4_tl+C4/tau4*f4_tl
  # fire flux from SOM -> atmosphere
  fire5=C5/tau5*f5
  fire5_tl=C5_tl/tau5*f5-fire5/tau5*tau5_tl+C5/tau5*f4_tl
  # total fire flux
  FireC=fire1+fire2+fire3+fire4+fire5
  FireC_tl=fire1_tl+fire2_tl+fire3_tl+fire4_tl+fire5_tl
  # 2. Nitrogen
  # N:C ratio of litter (Eq. 14 without multiplying C4)
  s4=(s1*(1.-eps1)*mort1+s2*(1.-eps2)*mort2)/(mort1+mort2)
  s4_tl=((s1_tl*(1.-eps1)-s1*eps1_tl)*mort1+s1*(1.-eps1)*mort1_tl\
        +(s2_tl*(1.-eps2)-s2*eps2_tl)*mort2+s2*(1.-eps2)*mort2_tl)/(mort1+mort2)\
        -s4/(mort1+mort2)*(mort1_tl+mort2_tl)
  # N demand for foliage NPP (first two items in Eq. E6)
  Fn1=s1*(mort1*(1.-eps1)+fire1)
  Fn1_tl=s1_tl*(mort1*(1.-eps1)+fire1)+s1*(mort1_tl*(1.-eps1)-mort1*eps1_tl+fire1_tl)
  # N demand for root NPP (third items in Eq. E6)
  Fn2=s2*(mort2*(1.-eps2)+fire2)
  Fn2_tl=s2_tl*(mort2*(1.-eps2)+fire2)+s2*(mort2_tl*(1.-eps2)-mort2*eps2_tl+fire2_tl)
  # N demand for wood NPP (forth items in Eq. E6)
  Fn3=s3*(mort3*(1.-eps3)+fire3)
  Fn3_tl=s3_tl*(mort3*(1.-eps3)+fire3)+s3*(mort3_tl*(1.-eps3)-mort3*eps3_tl+fire3_tl)
  # total N demand (=plant uptake)
  Fn=Fn1+Fn2+Fn3
  Fn_tl=Fn1_tl+Fn2_tl+Fn3_tl
  # fraction of Fn allocated to foliage
  bet1=Fn1/Fn
  bet1_tl=Fn1_tl/Fn-bet1/Fn*Fn_tl
  # fraction of Fn allocated to root
  bet2=Fn2/Fn
  bet2_tl=Fn2_tl/Fn-bet2/Fn*Fn_tl
  # fraction of Fn allocated to wood
  bet3=1.-bet1-bet2
  bet3_tl=-bet1_tl-bet2_tl
  # loss of N through fire
  FireN=s1*fire1+s2*fire2+s3*fire3+s4*fire4+s5*fire5
  FireN_tl=s1_tl*fire1+s1*fire1_tl+s2_tl*fire2+s2*fire2_tl+s3_tl*fire3+s3*fire3_tl+s4_tl*fire4+s4*fire4_tl+s5_tl*fire5+s5*fire5_tl
  # residence time of litter
  taun4=tau4
  taun4_tl=tau4_tl
  # leaching flux of N
  Fl=(1.-gfrac)*(Nd+Nfix-FireN)
  Fl_tl=-gfrac_tl*(Nd+Nfix-FireN)+(1.-gfrac)*(Nd_tl+Nfix_tl-FireN_tl)
  # denitrification flux
  Fg=Nd+Nfix-FireN-Fl
  Fg_tl=Nd_tl+Nfix_tl-FireN_tl-Fl_tl
  # inorganic N pool
  n6=Fl/leachingratio
  n6_tl=Fl_tl/leachingratio-n6/leachingratio*leachingratio_tl
  # N immobilization flux
  Nim=(s5-s4)*etac*C4/taun4+(s5-s3)*mort3*(1.-eps3)
  Nim_tl=(s5_tl-s4_tl)*etac*C4/taun4+(s5-s4)*(etac_tl*C4+etac*C4_tl)/taun4-Nim/taun4*taun4_tl\
        +(s5_tl-s3_tl)*mort3*(1.-eps3)+(s5-s3)*mort3_tl*(1.-eps3)-(s5-s3)*mort3*eps3_tl
  # N mineralization due to litter decomposition
  Nmin1=(1.-f4-etac)*s4*C4/taun4
  Nmin1_tl=(-f4_tl-etac_tl)*s4*C4/taun4+(1.-f4-etac)*(s4_tl*C4+s4*C4_tl)/taun4-Nmin1/taun4*taun4_tl
  # N mineralization due to SOM decomposition
  Nmin2=(1.-f5)*s5*C5/tau5
  Nmin2_tl=(-f5_tl)*s5*C5/tau5+(1.-f5)*(s5_tl*C5+s5*C5_tl)/tau5-Nmin2/tau5*tau5_tl
  # 3. Phosphorus
  # P:C ratio of litter (Eq. 25 without multiplying C4)
  l4=(l1*(1.-tht1)*mort1+l2*(1.-tht2)*mort2)/(mort1+mort2)
  l4_tl=((l1_tl*(1.-tht1)-l1*tht1_tl)*mort1+l1*(1.-tht1)*mort1_tl\
        +(l2_tl*(1.-tht2)-l2*tht2_tl)*mort2+l2*(1.-tht2)*mort2_tl)/(mort1+mort2)\
        -l4/(mort1+mort2)*(mort1_tl+mort2_tl)
  # P demand for foliage NPP (first two items in Eq. E17)
  Fp1=l1*(mort1*(1.-tht1)+fire1)
  Fp1_tl=l1_tl*(mort1*(1.-tht1)+fire1)+l1*(mort1_tl*(1.-tht1)-mort1*tht1_tl+fire1_tl)
  # P demand for root NPP (third items in Eq. E17)
  Fp2=l2*(mort2*(1.-tht2)+fire2)
  Fp2_tl=l2_tl*(mort2*(1.-tht2)+fire2)+l2*(mort2_tl*(1.-tht2)-mort2*tht2_tl+fire2_tl)
  # P demand for wood NPP (third items in Eq. E17)
  Fp3=l3*(mort3*(1.-tht3)+fire3)
  Fp3_tl=l3_tl*(mort3*(1.-tht3)+fire3)+l3*(mort3_tl*(1.-tht3)-mort3*tht3_tl+fire3_tl)
  # total P demand (=plant uptake)
  Fp=Fp1+Fp2+Fp3
  Fp_tl=Fp1_tl+Fp2_tl+Fp3_tl
  # fraction of Fp allocated to foliage
  phi1=Fp1/Fp
  phi1_tl=Fp1_tl/Fp-phi1/Fp*Fp_tl
  # fraction of Fp allocated to root
  phi2=Fp2/Fp
  phi2_tl=Fp2_tl/Fp-phi2/Fp*Fp_tl
  # fraction of Fp allocated to wood
  phi3=1.-phi1-phi2
  phi3_tl=-phi1_tl-phi2_tl
  # loss of P through fire
  FireP=l1*fire1+l2*fire2+l3*fire3+l4*fire4+l5*fire5
  FireP_tl=l1_tl*fire1+l1*fire1_tl+l2_tl*fire2+l2*fire2_tl+l3_tl*fire3+l3*fire3_tl+l4_tl*fire4+l4*fire4_tl+l5_tl*fire5+l5*fire5_tl
  # residence time of litter
  taup4=tau4
  taup4_tl=tau4_tl
  # leaching flux of P
  lossp6=leachingratio*dissfrac+fsorb
  lossp6_tl=leachingratio_tl*dissfrac+leachingratio*dissfrac_tl+fsorb_tl
  # labile soil P
  p6=(Pw+Pd-FireP*(1.-fashP))/lossp6
  p6_tl=(Pw_tl+Pd_tl-FireP_tl*(1.-fashP)+FireP*fashP_tl)/lossp6-p6/lossp6*lossp6_tl
  # P immobilization flux
  Pim=(l5-l4)*etac*C4/taup4+(l5-l3)*mort3*(1.-tht3)
  Pim_tl=(l5_tl-l4_tl)*etac*C4/taup4+(l5-l4)*(etac_tl*C4+etac*C4_tl)/taup4-Pim/taup4*taup4_tl\
        +(l5_tl-l3_tl)*mort3*(1.-tht3)+(l5-l3)*mort3_tl*(1.-tht3)-(l5-l3)*mort3*tht3_tl
  # P mineralization due to litter decomposition
  Pmin1=(1.-f4-etac)*l4*C4/taup4
  Pmin1_tl=(-f4_tl-etac_tl)*l4*C4/taup4+(1.-f4-etac)*(l4_tl*C4+l4*C4_tl)/taup4-Pmin1/taup4*taup4_tl
  # P mineralization due to SOM decomposition
  Pmin2=(1.-f5)*l5*C5/tau5
  Pmin2_tl=(-f5_tl)*l5*C5/tau5+(1.-f5)*(l5_tl*C5+l5*C5_tl)/tau5-Pmin2/tau5*tau5_tl

  # 4. additional variables potentially to be analyzed
  # N:P ratio in foliage
  NP1=s1/l1
  NP1_tl=s1_tl/l1-NP1/l1*l1_tl
  # N:P ratio in root
  NP2=s2/l2
  NP2_tl=s2_tl/l2-NP2/l2*l2_tl
  # N:P ratio in wood
  NP3=s3/l3
  NP3_tl=s3_tl/l3-NP3/l3*l3_tl
  # N:P ratio in litter
  NP4=s4/l4
  NP4_tl=s4_tl/l4-NP4/l4*l4_tl
  # N:P ratio in SOM
  NP5=s5/l5
  NP5_tl=s5_tl/l5-NP5/l5*l5_tl
  # residence time of N in foliage
  taun1=s1*C1/Fn1
  taun1_tl=(s1_tl*C1+s1*C1_tl)/Fn1-taun1/Fn1*Fn1_tl
  # residence time of N in vegetation
  taunbio=(s1*C1+s2*C2+s3*C3)/Fn
  taunbio_tl=(s1_tl*C1+s1*C1_tl+s2_tl*C2+s2*C2_tl+s3_tl*C3+s3*C3_tl)/Fn-taunbio/Fn*Fn_tl
  # residence time of inorganic soil N
  taun6=n6/(Nd+Nfix+Nmin1+Nmin2)
  taun6_tl=n6_tl/(Nd+Nfix+Nmin1+Nmin2)-taun6/(Nd+Nfix+Nmin1+Nmin2)*(Nd_tl+Nfix_tl+Nmin1_tl+Nmin2_tl)
  # NUE
  NUE=gpp/Fn
  NUE_tl=gpp_tl/Fn-NUE/Fn*Fn_tl
  # rseidence time of P in foliage
  taup1=l1*C1/Fp1
  taup1_tl=(l1_tl*C1+l1*C1_tl)/Fp1-taup1/Fp1*Fn1_tl
  # residencet time of P in vegetation
  taupbio=(l1*C1+l2*C2+l3*C3)/Fp
  taupbio_tl=(l1_tl*C1+l1*C1_tl+l2_tl*C2+l2*C2_tl+l3_tl*C3+l3*C3_tl)/Fp-taupbio/Fp*Fp_tl
  # residence time of labile soil P
  taup6=p6/(Pd+Pw+Pmin1+Pmin2+FireP*fashP)
  taup6_tl=p6_tl/(Pd+Pw+Pmin1+Pmin2+FireP*fashP)-taup6/(Pd+Pw+Pmin1+Pmin2+FireP*fashP)*(Pd_tl+Pw_tl+Pmin1_tl+Pmin2_tl+FireP_tl*fashP+FireP*fashP_tl)
  # PUE
  PUE=gpp/Fp
  PUE_tl=gpp_tl/Fp-PUE/Fp*Fp_tl
  # openness of N
  OpenN=(Nd+Nfix)/(gpp*gam1*s1+gpp*gam2*s2+gpp*gam3*s3)
  OpenN_tl=(Nd_tl+Nfix_tl)/(gpp*gam1*s1+gpp*gam2*s2+gpp*gam3*s3)\
          -OpenN/(gpp*gam1*s1+gpp*gam2*s2+gpp*gam3*s3)*(gpp_tl*(gam1*s1+gam2*s2+gam3*s3)+gpp*(gam1_tl*s1+gam1*s1_tl+gam2_tl*s2+gam2*s2_tl+gam3_tl*s3+gam3*s3_tl))
  # openness of P
  OpenP=(Pd+Pw)/(gpp*gam1*l1+gpp*gam2*l2+gpp*gam3*l3)
  OpenP_tl=(Pd_tl+Pw_tl)/(gpp*gam1*l1+gpp*gam2*l2+gpp*gam3*l3)\
          -OpenP/(gpp*gam1*l1+gpp*gam2*l2+gpp*gam3*l3)*(gpp_tl*(gam1*l1+gam2*l2+gam3*l3)+gpp*(gam1_tl*l1+gam1*l1_tl+gam2_tl*l2+gam2*l2_tl+gam3_tl*l3+gam3*l3_tl))
  # residence time of N in ecosystem
  RTN=(s1*C1+s2*C2+s3*C3+s4*C4+s5*C5)/(Nd+Nfix)
  RTN_tl=(s1_tl*C1+s1*C1_tl+s2_tl*C2+s2*C2_tl+s3_tl*C3+s3*C3_tl+s4_tl*C4+s4*C4_tl+s5_tl*C5+s5*C5_tl)/(Nd+Nfix)-RTN/(Nd+Nfix)*(Nd_tl+Nfix_tl)
  # residence time of P in ecosystem
  RTP=(l1*C1+l2*C2+l3*C3+l4*C4+l5*C5)/(Pd+Pw)
  RTP_tl=(l1_tl*C1+l1*C1_tl+l2_tl*C2+l2*C2_tl+l3_tl*C3+l3*C3_tl+l4_tl*C4+l4*C4_tl+l5_tl*C5+l5*C5_tl)/(Pd+Pw)-RTP/(Pd+Pw)*(Pd_tl+Pw_tl)

  # build sensitivity matrix by connecting sensitivity of each output to all inputs (X_tl) together
  # full sensitivity matrix: dimension is # of outputs x # of inputs
  # output variables ("X_tl" represent sensitivity):
  # C1-5  : C pool of foliage, root, wood, litter, SOM
  # Fn     : total N uptake by plants
  # s4     : litter N pool
  # taun4  : residence time of litter N
  # bet1-3 : fraction of Fn allocated to foliage, root, wood
  # n6     : inorganic soil N pool
  # Nim    : N immobilization
  # Nmin1  : N mineralization from litter decomposition
  # Nmin2  : N mineralization from SOM decomposition
  # Fp     : total P uptake by plants
  # l4     : litter P pool
  # taup4  : residence time of litter P
  # phi1-3 : fraction of Fp allocated to foliage, root, wood
  # p6     : labile soil P pool
  # Pim    : P immobilization
  # Pmin1  : P mineralization from litter decomposition
  # Pmin2  : P mineralization from SOM decomposition
  # FireC  : C loss (from foliage + root + wood + litter + SOM) through fire
  # FireN  : N loss (from foliage + root + wood + litter + SOM) through fire
  # FireP  : P loss (from foliage + root + wood + litter + SOM) through fire
  # NP1-5  : N:P ratio in foliage, root, wood, litter, SOM
  # taun1  : residence time of N in foliage
  # taunbio: residence time of N in plants
  # taun6  : residence time of inorganic soil N pool
  # taupbio: residence time of P in plants
  # taup6  : residence time of labile soil P pool
  # NUE    : nitrogen use efficiency
  # PUE    : nitrogen use efficiency
  # OpenN  : openness of N
  # OpenP  : openness of P
  # RTN    : residence of N in ecosystem
  # RTP    : residence of P in ecosystem
  H=np.array([C1_tl,C2_tl,C3_tl,C4_tl,C5_tl,Fn_tl,s4_tl,taun4_tl,bet1_tl,bet2_tl,bet3_tl,n6_tl,Nim_tl,Nmin1_tl,Nmin2_tl,\
              Fp_tl,l4_tl,taup4_tl,phi1_tl,phi2_tl,phi3_tl,p6_tl,Pim_tl,Pmin1_tl,Pmin2_tl,\
              FireC_tl,FireN_tl,FireP_tl,NP1_tl,NP2_tl,NP3_tl,NP4_tl,NP5_tl,\
              taun1_tl,taunbio_tl,taun6_tl,taup1_tl,taupbio_tl,taup6_tl,NUE_tl,PUE_tl,OpenN_tl,OpenP_tl,RTN_tl,RTP_tl])
  # covariance matrix of outputs by error propogation
  sigma=np.dot(np.dot(H,B),H.T)

  # return results, sensitivity matrix, and uncertainties, see above the meaning of different variables
  return np.array([C1,C2,C3,C4,C5,Fn,s4,taun4,bet1,bet2,bet3,n6,Nim,Nmin1,Nmin2,\
                   Fp,l4,taup4,phi1,phi2,phi3,p6,Pim,Pmin1,Pmin2,\
                   FireC,FireN,FireP,NP1,NP2,NP3,NP4,NP5,\
                   taun1,taunbio,taun6,taup1,taupbio,taup6,NUE,PUE,OpenN,OpenP,RTN,RTP]),H,sigma
         
def grass(X,X_tl,B):
"""
 X   : input data (mean value)
 X_tl: ancillary variable to compute tagent linear of the code
 B   : variances of input data
"""
  # read mean values from input
  # gpp    : GPP; 
  # gam1-3 : GPP allocation fraction to foliage, root, wood (= NPP allocation fraction*NPP/GPP)
  # etac   : fraction of litter outflux that enters SOM
  # tau1-5 : residence times of foliage, root, wood, litter, SOM
  # f1-5   : fraction of fire to total outgoing flux from foliage, root, wood, litter, SOM
  # s1-5   : N:C ratio of foliage, root, wood, SOM
  # eps1-3 : N resportion coefficients of foliage, root (=0), wood (=0)
  # Nd     : N deposition
  # Nfix   : BNF
  # leachingratio :  leaching ratio (runoff/soil-water-content)
  # gfrac  : fraction of denitrification to the total loss of N (denitrification + leaching)
  # l1-5   : P:C ratio of foliage, root, wood, SOM
  # tht1-3 : P resportion coefficients of foliage, root (=0), wood (=0)
  # Pw     : release of P from rock weathering
  # Pd     : P deposition
  # dissfrac : fraction of labile soil P that is dissolved in the soil water
  # fsorb  : transform rate from labile soil P to occluded P
  # npp    : NPP
  # fashP  : fraction of P retained in the residual ash during a fire (=0.75) 
  gpp,gam1,gam2,gam3,etac=X[:5]
  tau1,tau2,tau3,tau4,tau5=X[5:10]
  f1,f2,f3,f4,f5=X[10:15]
  s1,s2,s3,s5=X[15:19]
  eps1,eps2,eps3=X[19:22]
  Nd,Nfix=X[22:24]
  leachingratio=X[24]
  gfrac=X[25]
  l1,l2,l3,l5=X[26:30]
  tht1,tht2,tht3=X[30:33]
  Pw,Pd=X[33:35]
  dissfrac=X[35]
  fsorb=X[36]
  npp=X[37]
  fashP=X[38]
  # ancillary variable for tangent-linear computation of the code
  gpp_tl,gam1_tl,gam2_tl,gam3_tl,etac_tl=X_tl[:5]
  tau1_tl,tau2_tl,tau3_tl,tau4_tl,tau5_tl=X_tl[5:10]
  f1_tl,f2_tl,f3_tl,f4_tl,f5_tl=X_tl[10:15]
  s1_tl,s2_tl,s3_tl,s5_tl=X_tl[15:19]
  eps1_tl,eps2_tl,eps3_tl=X_tl[19:22]
  Nd_tl,Nfix_tl=X_tl[22:24]
  leachingratio_tl=X_tl[24]
  gfrac_tl=X_tl[25]
  l1_tl,l2_tl,l3_tl,l5_tl=X_tl[26:30]
  tht1_tl,tht2_tl,tht3_tl=X_tl[30:33]
  Pw_tl,Pd_tl=X_tl[33:35]
  dissfrac_tl=X_tl[35]
  fsorb_tl=X_tl[36]
  npp_tl=X_tl[37]
  fashP_tl=X_tl[38]

  # 1. C cycle
  # Steady-state (SS) foliage C pool (Eq. E1)
  C1=gpp*gam1*tau1
  C1_tl=gpp_tl*gam1*tau1+gpp*gam1_tl*tau1+gpp*gam1*tau1_tl
  # SS root C pool (Eq. E2)
  C2=gpp*gam2*tau2
  C2_tl=gpp_tl*gam2*tau2+gpp*gam2_tl*tau2+gpp*gam2*tau2_tl
  # SS wood C pool (=0) for grassland grid cell
  C3=0.
  C3_tl=0.*gpp_tl
  # SS litter C pool (Eq. E4)
  C4=gpp*(gam1*(1.-f1)+gam2*(1.-f2))*tau4
  C4_tl=gpp_tl*(gam1*(1.-f1)+gam2*(1.-f2))*tau4+gpp*(gam1_tl*(1.-f1)-gam1*f1_tl+gam2_tl*(1.-f2)-gam2*f2_tl)*tau4+gpp*(gam1+gam2)*tau4_tl
  # SS SOM C pool (Eq. E5)
  C5=etac*C4/tau4*tau5
  C5_tl=((etac_tl*C4+etac*C4_tl)/tau4-etac*C4/tau4**2*tau4_tl)*tau5+etac*C4/tau4*tau5_tl
  # NPP
  Fc=npp
  Fc_tl=npp_tl
  # mortality of foliage -> litter
  mort1=gpp*gam1*(1.-f1)
  mort1_tl=(gpp_tl*gam1+gpp*gam1_tl)*(1.-f1)-gpp*gam1*f1_tl
  # mortality of root -> litter
  mort2=gpp*gam2*(1.-f2)
  mort2_tl=(gpp_tl*gam2+gpp*gam2_tl)*(1.-f2)-gpp*gam2*f2_tl
  # mortality of wood -> SOM
  mort3=0.
  mort3_tl=0.*gpp_tl
  # fire flux from foliage -> atmosphere
  fire1=gpp*gam1*f1
  fire1_tl=gpp_tl*gam1*f1+gpp*gam1_tl*f1+gpp*gam1*f1_tl
  # fire flux from root -> atmosphere
  fire2=gpp*gam2*f2
  fire2_tl=gpp_tl*gam2*f2+gpp*gam2_tl*f2+gpp*gam2*f2_tl
  # fire flux from wood -> atmosphere
  fire3=0.
  fire3_tl=0.*gpp_tl
  # fire flux from litter -> atmosphere
  fire4=C4/tau4*f4
  fire4_tl=C4_tl/tau4*f4-fire4/tau4*tau4_tl+C4/tau4*f4_tl
  # fire flux from SOM -> atmosphere
  fire5=C5/tau5*f5
  fire5_tl=C5_tl/tau5*f5-fire5/tau5*tau5_tl+C5/tau5*f4_tl
  # total fire flux
  FireC=fire1+fire2+fire4+fire5
  FireC_tl=fire1_tl+fire2_tl+fire4_tl+fire5_tl
  # 2. Nitrogen
  # N:C ratio of litter (Eq. 14 without multiplying C4)
  s4=(s1*(1.-eps1)*mort1+s2*(1.-eps2)*mort2)/(mort1+mort2)
  s4_tl=((s1_tl*(1.-eps1)-s1*eps1_tl)*mort1+s1*(1.-eps1)*mort1_tl\
        +(s2_tl*(1.-eps2)-s2*eps2_tl)*mort2+s2*(1.-eps2)*mort2_tl)/(mort1+mort2)\
        -s4/(mort1+mort2)*(mort1_tl+mort2_tl)
  # N demand for foliage NPP (first two items in Eq. E6)
  Fn1=s1*(mort1*(1.-eps1)+fire1)
  Fn1_tl=s1_tl*(mort1*(1.-eps1)+fire1)+s1*(mort1_tl*(1.-eps1)-mort1*eps1_tl+fire1_tl)
  # N demand for root NPP (third items in Eq. E6)
  Fn2=s2*(mort2*(1.-eps2)+fire2)
  Fn2_tl=s2_tl*(mort2*(1.-eps2)+fire2)+s2*(mort2_tl*(1.-eps2)-mort2*eps2_tl+fire2_tl)
  # N demand for wood NPP (=0) for grassland grid cell
  Fn3=0.
  Fn3_tl=0.*s3_tl
  # total N demand (=plant uptake)
  Fn=Fn1+Fn2
  Fn_tl=Fn1_tl+Fn2_tl
  # fraction of Fn allocated to foliage
  bet1=Fn1/Fn
  bet1_tl=Fn1_tl/Fn-bet1/Fn*Fn_tl
  # fraction of Fn allocated to root
  bet2=1.-bet1
  bet2_tl=-bet1_tl
  # fraction of Fn allocated to wood
  bet3=1.-bet1-bet2
  bet3_tl=-bet1_tl-bet2_tl
  # loss of N through fire
  FireN=s1*fire1+s2*fire2+s4*fire4+s5*fire5
  FireN_tl=s1_tl*fire1+s1*fire1_tl+s2_tl*fire2+s2*fire2_tl+s4_tl*fire4+s4*fire4_tl+s5_tl*fire5+s5*fire5_tl
  # residence time of litter
  taun4=tau4
  taun4_tl=tau4_tl
  # leaching flux of N
  Fl=(1.-gfrac)*(Nd+Nfix-FireN)
  Fl_tl=-gfrac_tl*(Nd+Nfix-FireN)+(1.-gfrac)*(Nd_tl+Nfix_tl-FireN_tl)
  # denitrification flux
  Fg=Nd+Nfix-FireN-Fl
  Fg_tl=Nd_tl+Nfix_tl-FireN_tl-Fl_tl
  # inorganic N pool
  n6=Fl/leachingratio
  n6_tl=Fl_tl/leachingratio-n6/leachingratio*leachingratio_tl
  # N immobilization flux
  Nim=(s5-s4)*etac*C4/taun4
  Nim_tl=(s5_tl-s4_tl)*etac*C4/taun4+(s5-s4)*(etac_tl*C4+etac*C4_tl)/taun4-Nim/taun4*taun4_tl
  # N mineralization due to litter decomposition
  Nmin1=(1.-f4-etac)*s4*C4/taun4
  Nmin1_tl=(-f4_tl-etac_tl)*s4*C4/taun4+(1.-f4-etac)*(s4_tl*C4+s4*C4_tl)/taun4-Nmin1/taun4*taun4_tl
  # N mineralization due to SOM decomposition
  Nmin2=(1.-f5)*s5*C5/tau5
  Nmin2_tl=(-f5_tl)*s5*C5/tau5+(1.-f5)*(s5_tl*C5+s5*C5_tl)/tau5-Nmin2/tau5*tau5_tl
  # 3. Phosphorus
  # P:C ratio of litter (Eq. 25 without multiplying C4)
  l4=(l1*(1.-tht1)*mort1+l2*(1.-tht2)*mort2)/(mort1+mort2)
  l4_tl=((l1_tl*(1.-tht1)-l1*tht1_tl)*mort1+l1*(1.-tht1)*mort1_tl\
        +(l2_tl*(1.-tht2)-l2*tht2_tl)*mort2+l2*(1.-tht2)*mort2_tl)/(mort1+mort2)\
        -l4/(mort1+mort2)*(mort1_tl+mort2_tl)
  # P demand for foliage NPP (first two items in Eq. E17)
  Fp1=l1*(mort1*(1.-tht1)+fire1)
  Fp1_tl=l1_tl*(mort1*(1.-tht1)+fire1)+l1*(mort1_tl*(1.-tht1)-mort1*tht1_tl+fire1_tl)
  # P demand for root NPP (third items in Eq. E17)
  Fp2=l2*(mort2*(1.-tht2)+fire2)
  Fp2_tl=l2_tl*(mort2*(1.-tht2)+fire2)+l2*(mort2_tl*(1.-tht2)-mort2*tht2_tl+fire2_tl)
  # P demand for wood NPP (=0) for grassland grid cell
  Fp3=0.
  Fp3_tl=0.*l3_tl
  # total P demand (=plant uptake)
  Fp=Fp1+Fp2
  Fp_tl=Fp1_tl+Fp2_tl
  # fraction of Fp allocated to foliage
  phi1=Fp1/Fp
  phi1_tl=Fp1_tl/Fp-phi1/Fp*Fp_tl
  # fraction of Fp allocated to root
  phi2=1.-phi1
  phi2_tl=-phi1_tl
  # fraction of Fp allocated to wood
  phi3=1.-phi1-phi2
  phi3_tl=-phi1_tl-phi2_tl
  # loss of P through fire
  FireP=l1*fire1+l2*fire2+l4*fire4+l5*fire5
  FireP_tl=l1_tl*fire1+l1*fire1_tl+l2_tl*fire2+l2*fire2_tl+l4_tl*fire4+l4*fire4_tl+l5_tl*fire5+l5*fire5_tl
  # residence time of litter
  taup4=tau4
  taup4_tl=tau4_tl
  # leaching flux of P
  lossp6=leachingratio*dissfrac+fsorb
  lossp6_tl=leachingratio_tl*dissfrac+leachingratio*dissfrac_tl+fsorb_tl
  # labile soil P
  p6=(Pw+Pd-FireP*(1.-fashP))/lossp6
  p6_tl=(Pw_tl+Pd_tl-FireP_tl*(1.-fashP)+FireP*fashP_tl)/lossp6-p6/lossp6*lossp6_tl
  # P immobilization flux
  Pim=(l5-l4)*etac*C4/taup4
  Pim_tl=(l5_tl-l4_tl)*etac*C4/taup4+(l5-l4)*(etac_tl*C4+etac*C4_tl)/taup4-Pim/taup4*taup4_tl
  # P mineralization due to litter decomposition
  Pmin1=(1.-f4-etac)*l4*C4/taup4
  Pmin1_tl=(-f4_tl-etac_tl)*l4*C4/taup4+(1.-f4-etac)*(l4_tl*C4+l4*C4_tl)/taup4-Pmin1/taup4*taup4_tl
  # P mineralization due to SOM decomposition
  Pmin2=(1.-f5)*l5*C5/tau5
  Pmin2_tl=(-f5_tl)*l5*C5/tau5+(1.-f5)*(l5_tl*C5+l5*C5_tl)/tau5-Pmin2/tau5*tau5_tl

  # 4. additional variables potentially to be analyzed
  # N:P ratio in foliage
  NP1=s1/l1
  NP1_tl=s1_tl/l1-NP1/l1*l1_tl
  # N:P ratio in root
  NP2=s2/l2
  NP2_tl=s2_tl/l2-NP2/l2*l2_tl
  # N:P ratio in wood
  NP3=0.
  NP3_tl=0.*s3_tl
  # N:P ratio in litter
  NP4=s4/l4
  NP4_tl=s4_tl/l4-NP4/l4*l4_tl
  # N:P ratio in SOM
  NP5=s5/l5
  NP5_tl=s5_tl/l5-NP5/l5*l5_tl
  # residence time of N in foliage
  taun1=s1*C1/Fn1
  taun1_tl=(s1_tl*C1+s1*C1_tl)/Fn1-taun1/Fn1*Fn1_tl
  # residence time of N in vegetation
  taunbio=(s1*C1+s2*C2)/Fn
  taunbio_tl=(s1_tl*C1+s1*C1_tl+s2_tl*C2+s2*C2_tl)/Fn-taunbio/Fn*Fn_tl
  # residence time of inorganic soil N
  taun6=n6/(Nd+Nfix+Nmin1+Nmin2)
  taun6_tl=n6_tl/(Nd+Nfix+Nmin1+Nmin2)-taun6/(Nd+Nfix+Nmin1+Nmin2)*(Nd_tl+Nfix_tl+Nmin1_tl+Nmin2_tl)
  # NUE
  NUE=gpp/Fn
  NUE_tl=gpp_tl/Fn-NUE/Fn*Fn_tl
  # rseidence time of P in foliage
  taup1=l1*C1/Fp1
  taup1_tl=(l1_tl*C1+l1*C1_tl)/Fp1-taup1/Fp1*Fn1_tl
  # residencet time of P in vegetation
  taupbio=(l1*C1+l2*C2)/Fp
  taupbio_tl=(l1_tl*C1+l1*C1_tl+l2_tl*C2+l2*C2_tl)/Fp-taupbio/Fp*Fp_tl
  # residence time of labile soil P
  taup6=p6/(Pd+Pw+Pmin1+Pmin2+FireP*fashP)
  taup6_tl=p6_tl/(Pd+Pw+Pmin1+Pmin2+FireP*fashP)-taup6/(Pd+Pw+Pmin1+Pmin2+FireP*fashP)*(Pd_tl+Pw_tl+Pmin1_tl+Pmin2_tl+FireP_tl*fashP+FireP*fashP_tl)
  # PUE
  PUE=gpp/Fp
  PUE_tl=gpp_tl/Fp-PUE/Fp*Fp_tl
  # openness of N
  OpenN=(Nd+Nfix)/(gpp*gam1*s1+gpp*gam2*s2)
  OpenN_tl=(Nd_tl+Nfix_tl)/(gpp*gam1*s1+gpp*gam2*s2)\
          -OpenN/(gpp*gam1*s1+gpp*gam2*s2)*(gpp_tl*(gam1*s1+gam2*s2)+gpp*(gam1_tl*s1+gam1*s1_tl+gam2_tl*s2+gam2*s2_tl))
  # openness of P
  OpenP=(Pd+Pw)/(gpp*gam1*l1+gpp*gam2*l2)
  OpenP_tl=(Pd_tl+Pw_tl)/(gpp*gam1*l1+gpp*gam2*l2)\
          -OpenP/(gpp*gam1*l1+gpp*gam2*l2)*(gpp_tl*(gam1*l1+gam2*l2)+gpp*(gam1_tl*l1+gam1*l1_tl+gam2_tl*l2+gam2*l2_tl))
  # residence time of N in ecosystem
  RTN=(s1*C1+s2*C2+s4*C4+s5*C5)/(Nd+Nfix)
  RTN_tl=(s1_tl*C1+s1*C1_tl+s2_tl*C2+s2*C2_tl+s4_tl*C4+s4*C4_tl+s5_tl*C5+s5*C5_tl)/(Nd+Nfix)-RTN/(Nd+Nfix)*(Nd_tl+Nfix_tl)
  # residence time of P in ecosystem
  RTP=(l1*C1+l2*C2+l4*C4+l5*C5)/(Pd+Pw)
  RTP_tl=(l1_tl*C1+l1*C1_tl+l2_tl*C2+l2*C2_tl+l4_tl*C4+l4*C4_tl+l5_tl*C5+l5*C5_tl)/(Pd+Pw)-RTP/(Pd+Pw)*(Pd_tl+Pw_tl)

  # build sensitivity matrix by connecting sensitivity of each output to all inputs (X_tl) together
  # full sensitivity matrix: dimension is # of outputs x # of inputs
  # output variables ("X_tl" represent sensitivity):
  # C1-5  : C pool of foliage, root, wood, litter, SOM
  # Fn     : total N uptake by plants
  # s4     : litter N pool
  # taun4  : residence time of litter N
  # bet1-3 : fraction of Fn allocated to foliage, root, wood
  # n6     : inorganic soil N pool
  # Nim    : N immobilization
  # Nmin1  : N mineralization from litter decomposition
  # Nmin2  : N mineralization from SOM decomposition
  # Fp     : total P uptake by plants
  # l4     : litter P pool
  # taup4  : residence time of litter P
  # phi1-3 : fraction of Fp allocated to foliage, root, wood
  # p6     : labile soil P pool
  # Pim    : P immobilization
  # Pmin1  : P mineralization from litter decomposition
  # Pmin2  : P mineralization from SOM decomposition
  # FireC  : C loss (from foliage + root + wood + litter + SOM) through fire
  # FireN  : N loss (from foliage + root + wood + litter + SOM) through fire
  # FireP  : P loss (from foliage + root + wood + litter + SOM) through fire
  # NP1-5  : N:P ratio in foliage, root, wood, litter, SOM
  # taun1  : residence time of N in foliage
  # taunbio: residence time of N in plants
  # taun6  : residence time of inorganic soil N pool
  # taupbio: residence time of P in plants
  # taup6  : residence time of labile soil P pool
  # NUE    : nitrogen use efficiency
  # PUE    : nitrogen use efficiency
  # OpenN  : openness of N
  # OpenP  : openness of P
  # RTN    : residence of N in ecosystem
  # RTP    : residence of P in ecosystem
  H=np.array([C1_tl,C2_tl,C3_tl,C4_tl,C5_tl,Fn_tl,s4_tl,taun4_tl,bet1_tl,bet2_tl,bet3_tl,n6_tl,Nim_tl,Nmin1_tl,Nmin2_tl,\
              Fp_tl,l4_tl,taup4_tl,phi1_tl,phi2_tl,phi3_tl,p6_tl,Pim_tl,Pmin1_tl,Pmin2_tl,\
              FireC_tl,FireN_tl,FireP_tl,NP1_tl,NP2_tl,NP3_tl,NP4_tl,NP5_tl,\
              taun1_tl,taunbio_tl,taun6_tl,taup1_tl,taupbio_tl,taup6_tl,NUE_tl,PUE_tl,OpenN_tl,OpenP_tl,RTN_tl,RTP_tl])
  # covariance matrix of outputs by error propogation
  sigma=np.dot(np.dot(H,B),H.T)

  # return results, sensitivity matrix, and uncertainties, see above the meaning of different variables
  return np.array([C1,C2,C3,C4,C5,Fn,s4,taun4,bet1,bet2,bet3,n6,Nim,Nmin1,Nmin2,\
                   Fp,l4,taup4,phi1,phi2,phi3,p6,Pim,Pmin1,Pmin2,\
                   FireC,FireN,FireP,NP1,NP2,NP3,NP4,NP5,\
                   taun1,taunbio,taun6,taup1,taupbio,taup6,NUE,PUE,OpenN,OpenP,RTN,RTP]),H,sigma

